
from distutils.core import setup

setup(
    name='PyMemoize',
    version='1.0.3',
    description='Simple memoizing module.',
    url='http://github.com/mikeboers/PyMemoize',
    
    packages=['memoize'],
    
    author='Mike Boers',
    author_email='memoize@mikeboers.com',
    license='BSD-3',
    
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: BSD License',
        'Natural Language :: English',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 3',
        'Topic :: Software Development :: Libraries :: Python Modules',
    ],
    
)
